#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['jjowl']

package_data = \
{'': ['*']}

entry_points = \
{'console_scripts': ['jjtable2ttl = jjowl:main_table2ttl',
                     'jjtmd2ttl = jjowl:main_tmd2ttl',
                     'jjyaml2ttl = jjowl:main_yaml2ttl',
                     'owlcat = jjowl:mcat',
                     'owlclass = jjowl:mclass',
                     'owlexpandpp = jjowl:mexpandpp',
                     'owlgrep = jjowl:mgrep',
                     'owllabel2term = jjowl:mlabel2term',
                     'owlprops = jjowl:mprops',
                     'owlxdxf = jjowl:mxdxf']}

setup(name='jjowl',
      version='0.2.4',
      description='VERY EXPERIMENTAL :  module for opinied command line processing of OWL ontologies',
      author=None,
      author_email='JJoao <jj@di.uminho.pt>',
      url=None,
      packages=packages,
      package_data=package_data,
      entry_points=entry_points,
     )
